#include <stdio.h>
#include <stdlib.h>

#include "piece.h"
#include "inputspec.h"
#include "configuration.h"


extern InputSpec Spec;

/* Directional (x, y) offsets.  0 = right, 1 = down, 2 = left, 3 = up.
   This lets us use (dir + 2) % 4 to get a direction opposite to dir. */

int DX[] = { 1, 0, -1, 0 };
int DY[] = { 0, 1, 0, -1 };


/* Make a new piece of a given shape, in a given place. */

Piece *NewPiece(int whichShape, int l, int t)
{
Piece *result = (Piece *) malloc(sizeof(Piece));

	result->shapeIndex = whichShape;
	result->left = l;
	result->top = t;
	
	return result;
}


/* Make a new piece copied from an old one. */

Piece *CopyPiece(Piece *piece)
{
Piece *result = (Piece *) malloc (sizeof(Piece));

	*result = *piece;
	
	return result;
}


/* Is this the goal piece, in the goal position? */

int	PieceSolved(Piece *piece)
{
 	return IsGoalPiece(piece) && piece->left == Spec.goalX && piece->top == Spec.goalY;
}


/* Is this the goal piece? */

int	IsGoalPiece(Piece *piece)
{
	return piece->shapeIndex == Spec.goalIndex;
}


/* Does this piece occupy the spot (x, y)? */

int	PieceOccupies(Piece *piece, int x, int y)
{
	x -= piece->left;
	y -= piece->top;
	
	if (x < 0 || y < 0 || x >= MAXPIECEDIM || y >= MAXPIECEDIM)
		return 0;
		
	return Spec.shapes[piece->shapeIndex][x][y];
}


/* Can this piece slide in the specified direction, in the given configuration?
   We must make sure that the squares the piece would slide into are vacant.
   Those are the ones that are in the given direction from the squares the
   piece occupies, except for those such squares that the piece itself
   currently occupies. */

int	CanSlide(Piece *piece, int dir, struct Configuration *config)
{
	int x, y;
	
	for (x = piece->left; x <= piece->left + MAXPIECEDIM; ++x)
		for (y = piece->top; y <= piece->top + MAXPIECEDIM; ++y)
			if (PieceOccupies(piece, x, y) && !PieceOccupies(piece, x + DX[dir], y + DY[dir]) &&
			  ConfigOccupies(config, x + DX[dir], y + DY[dir]))
			  	return 0;
			  	
	return 1;
}		


/* Shift the piece. */

void Shift(Piece *piece, int dir)
{
	piece->left += DX[dir];
	piece->top += DY[dir];
}


/* Print out the piece info. */

void PrintPiece(Piece *piece)
{
	printf("%d %d %d\n", piece->shapeIndex, piece->left, piece->top);
}
	

/* Is this piece the same shape as and in the same location as another piece?
   Used for comparing configurations. */

int	SamePieceAs(Piece *piece1, Piece *piece2)
{
	return piece1->shapeIndex == piece2->shapeIndex && piece1->left == piece2->left && piece1->top == piece2->top;
}


/* Make a new piece list node. */

PieceList *NewPieceList(Piece *p, PieceList *pl)
{
PieceList *result = (PieceList *) malloc (sizeof(PieceList));

	result->piece = p;
	result->next = pl;
	
	return result;
}


/* Make a copy of an existing piece list. */

PieceList *CopyPieceList(PieceList *pl)
{
PieceList *result = (PieceList *) malloc (sizeof(PieceList));

	result->piece = CopyPiece(pl->piece);
	result->next = pl->next ? CopyPieceList(pl->next) : NULL;
	
	return result;
}


/* Dispose the contents of the list. */

void DeletePLContents(PieceList *pl)
{
	free(pl->piece);
	
	if (pl->next)
	{
		DeletePLContents(pl->next);
		free(pl->next);
	}
}


/* Print a PieceList.  The list is printed in order; this matches the order
   piece numbers are assigned when printing solutions. */

void PrintPieceList(PieceList *pl)
{
	PrintPiece(pl->piece);

	if (pl->next)
		PrintPieceList(pl->next);	
}


/* Find the pieceNumth piece in the list. */

Piece *PLIndex(PieceList *pl, int pieceNum)
{
	return pieceNum ? PLIndex(pl->next, pieceNum - 1) : pl->piece;
}
